
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "defines.h"
#include "files.h"



extern unsigned char *rom, *rom_usage;
extern int romsize, banksize, name_status, section_overwrite;
extern int pc_bank, pc_full, pc_slot, pc_slot_max;
extern struct object_file *obj_tmp;

int memory_file_id;



int mem_insert(int address, unsigned char data) {

  if (rom_usage[address]) {
    if (rom[address] == data)
      return SUCCEEDED;
    if (section_overwrite == ON)
      return SUCCEEDED;
    fprintf(stderr, "%s: MEM_INSERT: Overwrite at $%x [old: $%.1x new: $%.1x].\n", obj_tmp->name, address, rom[address], data);
  }

  rom_usage[address]++;
  rom[address] = data;

  return SUCCEEDED;
}


int mem_insert_gb_rom_header(int address, unsigned char data) {

  if (rom_usage[address]) {
    if (rom[address] == data)
      return SUCCEEDED;
    if (section_overwrite == ON)
      return SUCCEEDED;
    fprintf(stderr, "MEM_INSERT_GB_ROM_HEADER: Overwrite at $%x [old: $%.1x new: $%.1x].\n", address, rom[address], data);
  }

  rom_usage[address]++;
  rom[address] = data;

  return SUCCEEDED;
}


int mem_insert_ref(int address, unsigned char data) {

  if (address > romsize || address < 0) {
    fprintf(stderr, "MEM_INSERT_REF: Write (to %d) out of range (%d).\n", address, romsize);
    return FAILED;
  }

  if (rom_usage[address] > 1)
    fprintf(stderr, "MEM_INSERT_REF: Overwrite at $%x [old: $%.1x new: $%.1x].\n", address, rom[address], data);

  rom_usage[address]++;
  rom[address] = data;

  return SUCCEEDED;
}


int mem_insert_pc(unsigned char d) {

  if (pc_slot >= pc_slot_max) {
    fprintf(stderr, "%s: MEM_INSERT_PC: Overflow from slot.\n", get_file_name(memory_file_id));
    return FAILED;
  }
  if (pc_bank >= banksize) {
    fprintf(stderr, "%s: MEM_INSERT_PC: Overflow from bank.\n", get_file_name(memory_file_id));
    return FAILED;
  }
  if (pc_full >= romsize) {
    fprintf(stderr, "%s: MEM_INSERT_PC: Overflow from ROM file.\n", get_file_name(memory_file_id));
    return FAILED;
  }
  if (rom_usage[pc_full] != 0 && rom[pc_full] != d && section_overwrite == OFF)
    fprintf(stderr, "%s: MEM_INSERT_PC: Overwrite at $%x, old $%x new $%x\n", get_file_name(memory_file_id), pc_full, rom[pc_full], d);

  if (section_overwrite == OFF)
    rom_usage[pc_full]++;
  rom[pc_full] = d;
  pc_slot++;
  pc_bank++;
  pc_full++;

  return SUCCEEDED;
}


int mem_insert_pc_fix(unsigned char d) {

  if (pc_slot >= pc_slot_max) {
    fprintf(stderr, "%s: MEM_INSERT_PC_FIX: Overflow from slot.\n", get_file_name(memory_file_id));
    return FAILED;
  }
  if (pc_bank >= banksize) {
    fprintf(stderr, "%s: MEM_INSERT_PC_FIX: Overflow from bank.\n", get_file_name(memory_file_id));
    return FAILED;
  }
  if (pc_full >= romsize) {
    fprintf(stderr, "%s: MEM_INSERT_PC_FIX: Overflow from ROM file.\n", get_file_name(memory_file_id));
    return FAILED;
  }
  if (rom_usage[pc_full] > 1 && rom[pc_full] != d)
    fprintf(stderr, "%s: MEM_INSERT_PC_FIX: Overwrite at $%x, old $%x new $%x\n", get_file_name(memory_file_id), pc_full, rom[pc_full], d);

  rom_usage[pc_full]++;
  rom[pc_full] = d;
  pc_slot++;
  pc_bank++;
  pc_full++;

  return SUCCEEDED;
}
